From aa9ae5d3691d252187d031be70a080ccdcceea55 Mon Sep 17 00:00:00 2001
From: Sebastian Pop <spop@amazon.com>
Date: Tue, 10 Aug 2021 23:24:02 +0000
Subject: [PATCH] [arm64] port sse_{ss,pp} and ssd_s

         sse_pp[  4x4]	13.27x 	 6.83     	 90.66
         sse_ss[  4x4]	13.69x 	 7.04     	 96.40
         sse_pp[  8x8]	27.13x 	 12.64    	 343.05
         sse_ss[  8x8]	28.94x 	 12.71    	 367.81
  [i420] sse_pp[  4x4]	12.81x 	 7.08     	 90.61
  [i422] sse_pp[  4x8]	19.01x 	 9.45     	 179.58
         sse_pp[16x16]	37.61x 	 36.69    	 1379.72
         sse_ss[16x16]	39.30x 	 37.74    	 1482.94
  [i420] sse_pp[  8x8]	26.24x 	 13.10    	 343.74
  [i422] sse_pp[ 8x16]	33.71x 	 20.88    	 703.93
         sse_pp[32x32]	37.64x 	 145.17   	 5464.50
         sse_ss[32x32]	40.42x 	 145.60   	 5885.51
  [i420] sse_pp[16x16]	37.45x 	 36.93    	 1383.13
  [i422] sse_pp[16x32]	40.86x 	 67.42    	 2755.01
         sse_pp[64x64]	38.68x 	 566.98   	 21929.08
         sse_ss[64x64]	41.31x 	 568.08   	 23465.16
  [i420] sse_pp[32x32]	37.60x 	 145.30   	 5462.52
  [i422] sse_pp[32x64]	38.10x 	 286.07   	 10898.57
            ssd_s[4x4]	10.42x 	 6.41     	 66.77
            ssd_s[8x8]	23.52x 	 10.84    	 255.06
          ssd_s[16x16]	30.44x 	 33.62    	 1023.40
          ssd_s[32x32]	31.65x 	 128.51   	 4067.63
---
 source/common/CMakeLists.txt             |   2 +-
 source/common/aarch64/asm-primitives.cpp |  29 ++
 source/common/aarch64/pixel.h            |   4 +
 source/common/aarch64/ssd-a.S            | 498 +++++++++++++++++++++++
 4 files changed, 532 insertions(+), 1 deletion(-)
 create mode 100644 source/common/aarch64/ssd-a.S

diff --git a/source/common/CMakeLists.txt b/source/common/CMakeLists.txt
index f00aa7e..9099238 100644
--- a/source/common/CMakeLists.txt
+++ b/source/common/CMakeLists.txt
@@ -93,7 +93,7 @@ if(ENABLE_ASSEMBLY AND (ARM OR CROSS_COMPILE_ARM))
 #        enable_language(ASM)
 
         # add ARM assembly/intrinsic files here
-        set(A_SRCS asm.S mc-a.S sad-a.S pixel-util.S blockcopy8.S ipfilter8.S)
+        set(A_SRCS asm.S mc-a.S sad-a.S pixel-util.S ssd-a.S blockcopy8.S ipfilter8.S)
         set(VEC_PRIMITIVES)
 
         set(ARM_ASMS "${A_SRCS}" CACHE INTERNAL "ARM Assembly Sources")
diff --git a/source/common/aarch64/asm-primitives.cpp b/source/common/aarch64/asm-primitives.cpp
index db569e8..39aee44 100644
--- a/source/common/aarch64/asm-primitives.cpp
+++ b/source/common/aarch64/asm-primitives.cpp
@@ -515,6 +515,13 @@ void setupAssemblyPrimitives(EncoderPrimitives &p, int cpuMask)
         p.cu[BLOCK_16x16].cpy1Dto2D_shr = PFX(cpy1Dto2D_shr_16x16_neon);
         p.cu[BLOCK_32x32].cpy1Dto2D_shr = PFX(cpy1Dto2D_shr_32x32_neon);
 
+        // sse_ss
+        p.cu[BLOCK_4x4].sse_ss   = PFX(pixel_sse_ss_4x4_neon);
+        p.cu[BLOCK_8x8].sse_ss   = PFX(pixel_sse_ss_8x8_neon);
+        p.cu[BLOCK_16x16].sse_ss = PFX(pixel_sse_ss_16x16_neon);
+        p.cu[BLOCK_32x32].sse_ss = PFX(pixel_sse_ss_32x32_neon);
+        p.cu[BLOCK_64x64].sse_ss = PFX(pixel_sse_ss_64x64_neon);
+
         // pixel_sub_ps
         p.cu[BLOCK_4x4].sub_ps   = PFX(pixel_sub_ps_4x4_neon);
         p.cu[BLOCK_8x8].sub_ps   = PFX(pixel_sub_ps_8x8_neon);
@@ -538,6 +545,28 @@ void setupAssemblyPrimitives(EncoderPrimitives &p, int cpuMask)
         p.cu[BLOCK_16x16].calcresidual[NONALIGNED] = PFX(getResidual16_neon);
         p.cu[BLOCK_32x32].calcresidual[NONALIGNED] = PFX(getResidual32_neon);
 
+        // ssd_s
+        p.cu[BLOCK_4x4].ssd_s[NONALIGNED]   = PFX(pixel_ssd_s_4x4_neon);
+        p.cu[BLOCK_8x8].ssd_s[NONALIGNED]   = PFX(pixel_ssd_s_8x8_neon);
+        p.cu[BLOCK_16x16].ssd_s[NONALIGNED] = PFX(pixel_ssd_s_16x16_neon);
+        p.cu[BLOCK_32x32].ssd_s[NONALIGNED] = PFX(pixel_ssd_s_32x32_neon);
+
+        // sse_pp
+        p.cu[BLOCK_4x4].sse_pp   = PFX(pixel_sse_pp_4x4_neon);
+        p.cu[BLOCK_8x8].sse_pp   = PFX(pixel_sse_pp_8x8_neon);
+        p.cu[BLOCK_16x16].sse_pp = PFX(pixel_sse_pp_16x16_neon);
+        p.cu[BLOCK_32x32].sse_pp = PFX(pixel_sse_pp_32x32_neon);
+        p.cu[BLOCK_64x64].sse_pp = PFX(pixel_sse_pp_64x64_neon);
+
+        p.chroma[X265_CSP_I420].cu[BLOCK_420_4x4].sse_pp   = PFX(pixel_sse_pp_4x4_neon);
+        p.chroma[X265_CSP_I420].cu[BLOCK_420_8x8].sse_pp   = PFX(pixel_sse_pp_8x8_neon);
+        p.chroma[X265_CSP_I420].cu[BLOCK_420_16x16].sse_pp = PFX(pixel_sse_pp_16x16_neon);
+        p.chroma[X265_CSP_I420].cu[BLOCK_420_32x32].sse_pp = PFX(pixel_sse_pp_32x32_neon);
+        p.chroma[X265_CSP_I422].cu[BLOCK_422_4x8].sse_pp   = PFX(pixel_sse_pp_4x8_neon);
+        p.chroma[X265_CSP_I422].cu[BLOCK_422_8x16].sse_pp  = PFX(pixel_sse_pp_8x16_neon);
+        p.chroma[X265_CSP_I422].cu[BLOCK_422_16x32].sse_pp = PFX(pixel_sse_pp_16x32_neon);
+        p.chroma[X265_CSP_I422].cu[BLOCK_422_32x64].sse_pp = PFX(pixel_sse_pp_32x64_neon);
+
         // pixel_var
         p.cu[BLOCK_8x8].var   = PFX(pixel_var_8x8_neon);
         p.cu[BLOCK_16x16].var = PFX(pixel_var_16x16_neon);
diff --git a/source/common/aarch64/pixel.h b/source/common/aarch64/pixel.h
index dfe3e4a..f1bea34 100644
--- a/source/common/aarch64/pixel.h
+++ b/source/common/aarch64/pixel.h
@@ -130,9 +130,13 @@ void x265_sad_x4_64x48_neon(const pixel* fenc, const pixel* fref0, const pixel*
 void x265_sad_x4_64x64_neon(const pixel* fenc, const pixel* fref0, const pixel* fref1, const pixel* fref2, const pixel* fref3, intptr_t frefstride, int32_t* res);
 
 sse_t x265_pixel_sse_pp_4x4_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
+sse_t x265_pixel_sse_pp_4x8_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
 sse_t x265_pixel_sse_pp_8x8_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
+sse_t x265_pixel_sse_pp_8x16_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
 sse_t x265_pixel_sse_pp_16x16_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
+sse_t x265_pixel_sse_pp_16x32_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
 sse_t x265_pixel_sse_pp_32x32_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
+sse_t x265_pixel_sse_pp_32x64_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
 sse_t x265_pixel_sse_pp_64x64_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
 
 sse_t x265_pixel_sse_ss_4x4_neon(const int16_t* pix1, intptr_t stride_pix1, const int16_t* pix2, intptr_t stride_pix2);
diff --git a/source/common/aarch64/ssd-a.S b/source/common/aarch64/ssd-a.S
new file mode 100644
index 0000000..ac2d62d
--- /dev/null
+++ b/source/common/aarch64/ssd-a.S
@@ -0,0 +1,498 @@
+/*****************************************************************************
+ * Copyright (C) 2021 MulticoreWare, Inc
+ *
+ * Authors: Sebastian Pop <spop@amazon.com>
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
+ *
+ * This program is also available under a commercial proprietary license.
+ * For more information, contact us at license @ x265.com.
+ *****************************************************************************/
+
+#include "asm.S"
+
+.section .rodata
+
+.align 4
+
+.text
+
+.macro ret_v0_w0
+    trn2            v1.2d, v0.2d, v0.2d
+    add             v0.2s, v0.2s, v1.2s
+    addp            v0.2s, v0.2s, v0.2s
+    fmov            w0, s0
+    ret
+.endm
+
+function x265_pixel_sse_pp_4x4_neon
+    ld1             {v16.8b}, [x0], x1
+    ld1             {v17.8b}, [x2], x3
+    usubl           v1.8h, v16.8b, v17.8b
+    ld1             {v18.8b}, [x0], x1
+    smull           v0.4s, v1.4h, v1.4h
+    ld1             {v19.8b}, [x2], x3
+
+    usubl           v2.8h, v18.8b, v19.8b
+    ld1             {v20.8b}, [x0], x1
+    smlal           v0.4s, v2.4h, v2.4h
+    ld1             {v21.8b}, [x2], x3
+
+    usubl           v3.8h, v20.8b, v21.8b
+    ld1             {v22.8b}, [x0], x1
+    smlal           v0.4s, v3.4h, v3.4h
+    ld1             {v23.8b}, [x2], x3
+
+    usubl           v4.8h, v22.8b, v23.8b
+    smlal           v0.4s, v4.4h, v4.4h
+    ret_v0_w0
+endfunc
+
+function x265_pixel_sse_pp_4x8_neon
+    ld1             {v16.8b}, [x0], x1
+    ld1             {v17.8b}, [x2], x3
+    usubl           v1.8h, v16.8b, v17.8b
+    ld1             {v16.8b}, [x0], x1
+    smull           v0.4s, v1.4h, v1.4h
+    ld1             {v17.8b}, [x2], x3
+.rept 6
+    usubl           v1.8h, v16.8b, v17.8b
+    ld1             {v16.8b}, [x0], x1
+    smlal           v0.4s, v1.4h, v1.4h
+    ld1             {v17.8b}, [x2], x3
+.endr
+    usubl           v1.8h, v16.8b, v17.8b
+    smlal           v0.4s, v1.4h, v1.4h
+    ret_v0_w0
+endfunc
+
+function x265_pixel_sse_pp_8x8_neon
+    ld1             {v16.8b}, [x0], x1
+    ld1             {v17.8b}, [x2], x3
+    usubl           v1.8h, v16.8b, v17.8b
+    ld1             {v16.8b}, [x0], x1
+    smull           v0.4s, v1.4h, v1.4h
+    smlal2          v0.4s, v1.8h, v1.8h
+    ld1             {v17.8b}, [x2], x3
+
+.rept 6
+    usubl           v1.8h, v16.8b, v17.8b
+    ld1             {v16.8b}, [x0], x1
+    smlal           v0.4s, v1.4h, v1.4h
+    smlal2          v0.4s, v1.8h, v1.8h
+    ld1             {v17.8b}, [x2], x3
+.endr
+    usubl           v1.8h, v16.8b, v17.8b
+    smlal           v0.4s, v1.4h, v1.4h
+    smlal2          v0.4s, v1.8h, v1.8h
+    ret_v0_w0
+endfunc
+
+function x265_pixel_sse_pp_8x16_neon
+    ld1             {v16.8b}, [x0], x1
+    ld1             {v17.8b}, [x2], x3
+    usubl           v1.8h, v16.8b, v17.8b
+    ld1             {v16.8b}, [x0], x1
+    smull           v0.4s, v1.4h, v1.4h
+    smlal2          v0.4s, v1.8h, v1.8h
+    ld1             {v17.8b}, [x2], x3
+
+.rept 14
+    usubl           v1.8h, v16.8b, v17.8b
+    ld1             {v16.8b}, [x0], x1
+    smlal           v0.4s, v1.4h, v1.4h
+    smlal2          v0.4s, v1.8h, v1.8h
+    ld1             {v17.8b}, [x2], x3
+.endr
+    usubl           v1.8h, v16.8b, v17.8b
+    smlal           v0.4s, v1.4h, v1.4h
+    smlal2          v0.4s, v1.8h, v1.8h
+    ret_v0_w0
+endfunc
+
+function x265_pixel_sse_pp_16x16_neon
+    ld1             {v16.16b}, [x0], x1
+    ld1             {v17.16b}, [x2], x3
+    usubl           v1.8h, v16.8b, v17.8b
+    usubl2          v2.8h, v16.16b, v17.16b
+    ld1             {v16.16b}, [x0], x1
+    smull           v0.4s, v1.4h, v1.4h
+    smlal2          v0.4s, v1.8h, v1.8h
+    ld1             {v17.16b}, [x2], x3
+    smlal           v0.4s, v2.4h, v2.4h
+    smlal2          v0.4s, v2.8h, v2.8h
+
+.rept 14
+    usubl           v1.8h, v16.8b, v17.8b
+    usubl2          v2.8h, v16.16b, v17.16b
+    ld1             {v16.16b}, [x0], x1
+    smlal           v0.4s, v1.4h, v1.4h
+    smlal2          v0.4s, v1.8h, v1.8h
+    ld1             {v17.16b}, [x2], x3
+    smlal           v0.4s, v2.4h, v2.4h
+    smlal2          v0.4s, v2.8h, v2.8h
+.endr
+    usubl           v1.8h, v16.8b, v17.8b
+    usubl2          v2.8h, v16.16b, v17.16b
+    smlal           v0.4s, v1.4h, v1.4h
+    smlal2          v0.4s, v1.8h, v1.8h
+    smlal           v0.4s, v2.4h, v2.4h
+    smlal2          v0.4s, v2.8h, v2.8h
+    ret_v0_w0
+endfunc
+
+function x265_pixel_sse_pp_16x32_neon
+    mov             w12, #8
+    movi            v0.16b, #0
+    movi            v1.16b, #0
+.loop_sse_pp_16x32:
+    sub             w12, w12, #1
+.rept 4
+    ld1             {v16.16b}, [x0], x1
+    ld1             {v17.16b}, [x2], x3
+    usubl           v2.8h, v16.8b, v17.8b
+    usubl2          v3.8h, v16.16b, v17.16b
+    smlal           v0.4s, v2.4h, v2.4h
+    smlal2          v1.4s, v2.8h, v2.8h
+    smlal           v0.4s, v3.4h, v3.4h
+    smlal2          v1.4s, v3.8h, v3.8h
+.endr
+    cbnz            w12, .loop_sse_pp_16x32
+    add             v0.4s, v0.4s, v1.4s
+    ret_v0_w0
+endfunc
+
+function x265_pixel_sse_pp_32x32_neon
+    mov             w12, #8
+    movi            v0.16b, #0
+    movi            v1.16b, #0
+.loop_sse_pp_32:
+    sub             w12, w12, #1
+.rept 4
+    ld1             {v16.16b,v17.16b}, [x0], x1
+    ld1             {v18.16b,v19.16b}, [x2], x3
+    usubl           v2.8h, v16.8b, v18.8b
+    usubl2          v3.8h, v16.16b, v18.16b
+    usubl           v4.8h, v17.8b, v19.8b
+    usubl2          v5.8h, v17.16b, v19.16b
+    smlal           v0.4s, v2.4h, v2.4h
+    smlal2          v1.4s, v2.8h, v2.8h
+    smlal           v0.4s, v3.4h, v3.4h
+    smlal2          v1.4s, v3.8h, v3.8h
+    smlal           v0.4s, v4.4h, v4.4h
+    smlal2          v1.4s, v4.8h, v4.8h
+    smlal           v0.4s, v5.4h, v5.4h
+    smlal2          v1.4s, v5.8h, v5.8h
+.endr
+    cbnz            w12, .loop_sse_pp_32
+    add             v0.4s, v0.4s, v1.4s
+    ret_v0_w0
+endfunc
+
+function x265_pixel_sse_pp_32x64_neon
+    mov             w12, #16
+    movi            v0.16b, #0
+    movi            v1.16b, #0
+.loop_sse_pp_32x64:
+    sub             w12, w12, #1
+.rept 4
+    ld1             {v16.16b,v17.16b}, [x0], x1
+    ld1             {v18.16b,v19.16b}, [x2], x3
+    usubl           v2.8h, v16.8b, v18.8b
+    usubl2          v3.8h, v16.16b, v18.16b
+    usubl           v4.8h, v17.8b, v19.8b
+    usubl2          v5.8h, v17.16b, v19.16b
+    smlal           v0.4s, v2.4h, v2.4h
+    smlal2          v1.4s, v2.8h, v2.8h
+    smlal           v0.4s, v3.4h, v3.4h
+    smlal2          v1.4s, v3.8h, v3.8h
+    smlal           v0.4s, v4.4h, v4.4h
+    smlal2          v1.4s, v4.8h, v4.8h
+    smlal           v0.4s, v5.4h, v5.4h
+    smlal2          v1.4s, v5.8h, v5.8h
+.endr
+    cbnz            w12, .loop_sse_pp_32x64
+    add             v0.4s, v0.4s, v1.4s
+    ret_v0_w0
+endfunc
+
+function x265_pixel_sse_pp_64x64_neon
+    mov             w12, #16
+    movi            v0.16b, #0
+    movi            v1.16b, #0
+
+.loop_sse_pp_64:
+    sub             w12, w12, #1
+.rept 4
+    ld1             {v16.16b-v19.16b}, [x0], x1
+    ld1             {v20.16b-v23.16b}, [x2], x3
+
+    usubl           v2.8h, v16.8b, v20.8b
+    usubl2          v3.8h, v16.16b, v20.16b
+    usubl           v4.8h, v17.8b, v21.8b
+    usubl2          v5.8h, v17.16b, v21.16b
+    smlal           v0.4s, v2.4h, v2.4h
+    smlal2          v1.4s, v2.8h, v2.8h
+    smlal           v0.4s, v3.4h, v3.4h
+    smlal2          v1.4s, v3.8h, v3.8h
+    smlal           v0.4s, v4.4h, v4.4h
+    smlal2          v1.4s, v4.8h, v4.8h
+    smlal           v0.4s, v5.4h, v5.4h
+    smlal2          v1.4s, v5.8h, v5.8h
+
+    usubl           v2.8h, v18.8b, v22.8b
+    usubl2          v3.8h, v18.16b, v22.16b
+    usubl           v4.8h, v19.8b, v23.8b
+    usubl2          v5.8h, v19.16b, v23.16b
+    smlal           v0.4s, v2.4h, v2.4h
+    smlal2          v1.4s, v2.8h, v2.8h
+    smlal           v0.4s, v3.4h, v3.4h
+    smlal2          v1.4s, v3.8h, v3.8h
+    smlal           v0.4s, v4.4h, v4.4h
+    smlal2          v1.4s, v4.8h, v4.8h
+    smlal           v0.4s, v5.4h, v5.4h
+    smlal2          v1.4s, v5.8h, v5.8h
+.endr
+    cbnz            w12, .loop_sse_pp_64
+    add             v0.4s, v0.4s, v1.4s
+    ret_v0_w0
+endfunc
+
+function x265_pixel_sse_ss_4x4_neon
+    add             x1, x1, x1
+    add             x3, x3, x3
+
+    ld1             {v16.8b}, [x0], x1
+    ld1             {v17.8b}, [x2], x3
+    sub             v2.4h, v16.4h, v17.4h
+    ld1             {v16.8b}, [x0], x1
+    smull           v0.4s, v2.4h, v2.4h
+    ld1             {v17.8b}, [x2], x3
+
+    sub             v2.4h, v16.4h, v17.4h
+    ld1             {v16.8b}, [x0], x1
+    smlal           v0.4s, v2.4h, v2.4h
+    ld1             {v17.8b}, [x2], x3
+
+    sub             v2.4h, v16.4h, v17.4h
+    ld1             {v16.8b}, [x0], x1
+    smlal           v0.4s, v2.4h, v2.4h
+    ld1             {v17.8b}, [x2], x3
+
+    sub             v2.4h, v16.4h, v17.4h
+    smlal           v0.4s, v2.4h, v2.4h
+
+    ret_v0_w0
+endfunc
+
+function x265_pixel_sse_ss_8x8_neon
+    add             x1, x1, x1
+    add             x3, x3, x3
+    ld1             {v16.16b}, [x0], x1
+    ld1             {v17.16b}, [x2], x3
+    sub             v2.8h, v16.8h, v17.8h
+    smull           v0.4s, v2.4h, v2.4h
+    smull2          v1.4s, v2.8h, v2.8h
+.rept 7
+    ld1             {v16.16b}, [x0], x1
+    ld1             {v17.16b}, [x2], x3
+    sub             v2.8h, v16.8h, v17.8h
+    smlal           v0.4s, v2.4h, v2.4h
+    smlal2          v1.4s, v2.8h, v2.8h
+.endr
+    add             v0.4s, v0.4s, v1.4s
+    ret_v0_w0
+endfunc
+
+function x265_pixel_sse_ss_16x16_neon
+    add             x1, x1, x1
+    add             x3, x3, x3
+
+    mov             w12, #4
+    movi            v0.16b, #0
+    movi            v1.16b, #0
+.loop_sse_ss_16:
+    sub             w12, w12, #1
+.rept 4
+    ld1             {v16.16b,v17.16b}, [x0], x1
+    ld1             {v18.16b,v19.16b}, [x2], x3
+    sub             v2.8h, v16.8h, v18.8h
+    sub             v3.8h, v17.8h, v19.8h
+    smlal           v0.4s, v2.4h, v2.4h
+    smlal2          v1.4s, v2.8h, v2.8h
+    smlal           v0.4s, v3.4h, v3.4h
+    smlal2          v1.4s, v3.8h, v3.8h
+.endr
+    cbnz            w12, .loop_sse_ss_16
+    add             v0.4s, v0.4s, v1.4s
+    ret_v0_w0
+endfunc
+
+function x265_pixel_sse_ss_32x32_neon
+    add             x1, x1, x1
+    add             x3, x3, x3
+
+    mov             w12, #8
+    movi            v0.16b, #0
+    movi            v1.16b, #0
+.loop_sse_ss_32:
+    sub             w12, w12, #1
+.rept 4
+    ld1             {v16.16b-v19.16b}, [x0], x1
+    ld1             {v20.16b-v23.16b}, [x2], x3
+    sub             v2.8h, v16.8h, v20.8h
+    sub             v3.8h, v17.8h, v21.8h
+    sub             v4.8h, v18.8h, v22.8h
+    sub             v5.8h, v19.8h, v23.8h
+    smlal           v0.4s, v2.4h, v2.4h
+    smlal2          v1.4s, v2.8h, v2.8h
+    smlal           v0.4s, v3.4h, v3.4h
+    smlal2          v1.4s, v3.8h, v3.8h
+    smlal           v0.4s, v4.4h, v4.4h
+    smlal2          v1.4s, v4.8h, v4.8h
+    smlal           v0.4s, v5.4h, v5.4h
+    smlal2          v1.4s, v5.8h, v5.8h
+.endr
+    cbnz            w12, .loop_sse_ss_32
+    add             v0.4s, v0.4s, v1.4s
+    ret_v0_w0
+endfunc
+
+function x265_pixel_sse_ss_64x64_neon
+    add             x1, x1, x1
+    add             x3, x3, x3
+    sub             x1, x1, #64
+    sub             x3, x3, #64
+
+    mov             w12, #32
+    movi            v0.16b, #0
+    movi            v1.16b, #0
+.loop_sse_ss_64:
+    sub             w12, w12, #1
+.rept 2
+    ld1             {v16.16b-v19.16b}, [x0], #64
+    ld1             {v20.16b-v23.16b}, [x2], #64
+    sub             v2.8h, v16.8h, v20.8h
+    sub             v3.8h, v17.8h, v21.8h
+    sub             v4.8h, v18.8h, v22.8h
+    sub             v5.8h, v19.8h, v23.8h
+    smlal           v0.4s, v2.4h, v2.4h
+    smlal2          v1.4s, v2.8h, v2.8h
+    smlal           v0.4s, v3.4h, v3.4h
+    smlal2          v1.4s, v3.8h, v3.8h
+    smlal           v0.4s, v4.4h, v4.4h
+    smlal2          v1.4s, v4.8h, v4.8h
+    smlal           v0.4s, v5.4h, v5.4h
+    smlal2          v1.4s, v5.8h, v5.8h
+    ld1             {v16.16b-v19.16b}, [x0], x1
+    ld1             {v20.16b-v23.16b}, [x2], x3
+    sub             v2.8h, v16.8h, v20.8h
+    sub             v3.8h, v17.8h, v21.8h
+    sub             v4.8h, v18.8h, v22.8h
+    sub             v5.8h, v19.8h, v23.8h
+    smlal           v0.4s, v2.4h, v2.4h
+    smlal2          v1.4s, v2.8h, v2.8h
+    smlal           v0.4s, v3.4h, v3.4h
+    smlal2          v1.4s, v3.8h, v3.8h
+    smlal           v0.4s, v4.4h, v4.4h
+    smlal2          v1.4s, v4.8h, v4.8h
+    smlal           v0.4s, v5.4h, v5.4h
+    smlal2          v1.4s, v5.8h, v5.8h
+.endr
+    cbnz            w12, .loop_sse_ss_64
+    add             v0.4s, v0.4s, v1.4s
+    ret_v0_w0
+endfunc
+
+function x265_pixel_ssd_s_4x4_neon
+    add             x1, x1, x1
+    ld1             {v4.8b}, [x0], x1
+    ld1             {v5.8b}, [x0], x1
+    ld1             {v6.8b}, [x0], x1
+    ld1             {v7.8b}, [x0]
+    smull           v0.4s, v4.4h, v4.4h
+    smull           v1.4s, v5.4h, v5.4h
+    smlal           v0.4s, v6.4h, v6.4h
+    smlal           v1.4s, v7.4h, v7.4h
+    add             v0.4s, v0.4s, v1.4s
+    ret_v0_w0
+endfunc
+
+function x265_pixel_ssd_s_8x8_neon
+    add             x1, x1, x1
+    ld1             {v4.16b}, [x0], x1
+    ld1             {v5.16b}, [x0], x1
+    smull           v0.4s, v4.4h, v4.4h
+    smull2          v1.4s, v4.8h, v4.8h
+    smlal           v0.4s, v5.4h, v5.4h
+    smlal2          v1.4s, v5.8h, v5.8h
+.rept 3
+    ld1             {v4.16b}, [x0], x1
+    ld1             {v5.16b}, [x0], x1
+    smlal           v0.4s, v4.4h, v4.4h
+    smlal2          v1.4s, v4.8h, v4.8h
+    smlal           v0.4s, v5.4h, v5.4h
+    smlal2          v1.4s, v5.8h, v5.8h
+.endr
+    add             v0.4s, v0.4s, v1.4s
+    ret_v0_w0
+endfunc
+
+function x265_pixel_ssd_s_16x16_neon
+    add             x1, x1, x1
+    mov             w12, #4
+    movi            v0.16b, #0
+    movi            v1.16b, #0
+.loop_ssd_s_16:
+    sub             w12, w12, #1
+.rept 2
+    ld1             {v4.16b,v5.16b}, [x0], x1
+    ld1             {v6.16b,v7.16b}, [x0], x1
+    smlal           v0.4s, v4.4h, v4.4h
+    smlal2          v1.4s, v4.8h, v4.8h
+    smlal           v0.4s, v5.4h, v5.4h
+    smlal2          v1.4s, v5.8h, v5.8h
+    smlal           v0.4s, v6.4h, v6.4h
+    smlal2          v1.4s, v6.8h, v6.8h
+    smlal           v0.4s, v7.4h, v7.4h
+    smlal2          v1.4s, v7.8h, v7.8h
+.endr
+    cbnz            w12, .loop_ssd_s_16
+    add             v0.4s, v0.4s, v1.4s
+    ret_v0_w0
+endfunc
+
+function x265_pixel_ssd_s_32x32_neon
+    add             x1, x1, x1
+    mov             w12, #8
+    movi            v0.16b, #0
+    movi            v1.16b, #0
+.loop_ssd_s_32:
+    sub             w12, w12, #1
+.rept 4
+    ld1             {v4.16b-v7.16b}, [x0], x1
+    smlal           v0.4s, v4.4h, v4.4h
+    smlal2          v1.4s, v4.8h, v4.8h
+    smlal           v0.4s, v5.4h, v5.4h
+    smlal2          v1.4s, v5.8h, v5.8h
+    smlal           v0.4s, v6.4h, v6.4h
+    smlal2          v1.4s, v6.8h, v6.8h
+    smlal           v0.4s, v7.4h, v7.4h
+    smlal2          v1.4s, v7.8h, v7.8h
+.endr
+    cbnz            w12, .loop_ssd_s_32
+    add             v0.4s, v0.4s, v1.4s
+    ret_v0_w0
+endfunc
-- 
2.25.1

